#!/usr/bin/perl -w
use strict;
use diagnostics;


#
#
#        CS261 Security in Computer Systems
#        Homework 1
#
#
#        inanimateMail.pl : Reads in (potentially) malicious HTML and
#                           prints a sanitized version.
#
#
#        Acknowledgements: Extracts from
#                              "Parsing HTML with HTML::PARSER - 
#                               The Perl Journal, Spring 2000"
#        (www.foo.be/docs/tpj/issues/vol5_1/tpj0501-0003.html)
#
#
#        2011-09-14
#
#


package InanimateMail;

use base "HTML::Parser";
use File::Temp;


#
#        CONSTANTS
#


my $TRUE  = 1;
my $FALSE = 0;


# These tags are printed out verbatim (attributes are handled separately).
# Other tags are discarded.
my @PERMITTED_TAGS
    = qw (html
          head body
          p
          h1 h2 h3 h4
          div
          table hr tr td
          a
          i u b
          pre tt blockquote code
          dl dt dd
          li ol ul
          br
          font
          img);
my %PERMITTED_TAGS = ();

# These tags, which are used for printing preformatted text:
# 1) require slightly different sanitization
# 2) are converted to code tags for convenience
my %BLOCKQUOTE_TAGS
    = (
       'pre'        => 'code',
       'tt'         => 'code',
       'blockquote' => 'code',
       'code'       => 'code');

# Regular expressions, used for matching attribute values.
# The start () function will require that the regex match the whole string.
my $ALIGNMENTS = '(left|right|middle|center)';
my $COLORS =   '('
             . '(#[a-f0-9]{6})'
             . '|white|red|green|blue|yellow|pink|purple|grey|black'
             . ')';
my $NUMERIC = '[0-9]+';

# These attributes are printed as part of the tag, if the value matches
# the regex.
my %PERMITTED_ATTRIBUTES
    = (
       border      => $NUMERIC,
       width       => $NUMERIC,
       cellspacing => $NUMERIC,
       cellpadding => $NUMERIC,
       align       => $ALIGNMENTS,
       bgcolor     => $COLORS,
       color       => $COLORS,
       size        => $NUMERIC
      );

# Attributes for these specific tags are printed as plain (censored) text.
my %INLINED_ATTRIBUTES
    = (
       'a'   => 'href',
       'img' => 'src'
      );

# Characters not in this range are omitted.
my $PERMITTED_CHARS = "a-z0-9;:)(\' \t\n<>/,.?|^~!@#\$%^&*(){}_=+\"\Q[-\\]\E";

# These invisible tags are used to separate text (including potentially
# any scripts lurking around).
my $NOP          = '<code></code>';
# If we are already inside a code tag, we don't want to nest them.
my $NOP_FOR_CODE = '</code><code>';

# These specific HTML entities are allowed.
my @ENTITIES
    = qw (quot amp apos lt gt nbsp);
my %ENTITIES = ();


#
#        GLOBALS (yuck)
#


my %REJECTED_TAGS = ();
my %REJECTED_CHARS = ();

my @TAG_STACK = ();


#
#        FUNCTIONS
#


sub text ($$);
sub start ($$);
sub end ($$);
sub stripComments ($);
sub printOrdered ($);


# printCensored : Removes non-whitelisted characters, and inserts
#                 redundant tags between each character.
sub printCensored ($) {
    my ($text) = @_;

    my $censored = $text;
    $censored =~ s/[$PERMITTED_CHARS]//gi;
    foreach my $char (split //, $censored) {
        $REJECTED_CHARS {$char} ++;
    }

    $text =~ s/[^$PERMITTED_CHARS]//gi;

    my ($padded);

    if (   (@TAG_STACK > 0)
        && (defined $BLOCKQUOTE_TAGS {$TAG_STACK [-1]})) {
        my @text = split //, $text;
        $padded = join $NOP_FOR_CODE, @text;
        $padded =~ s/\n/<br\/>\n/g;
    } else {
        my @text = split //, $text;
        $padded = join $NOP, @text;
    }

    foreach my $escapedTag (keys %ENTITIES) {
        my $tag = $ENTITIES {$escapedTag};
        $padded =~ s/\Q$escapedTag\E/$tag/g;
    }

    print $padded;
}


# printUncensored : Prints the text verbatim.
#
#                   This function is used as a wrapper around 'print'
#                   for the convenience of identifying dangerous calls.
sub printUncensored ($) {
    my ($text) = @_;

    print $text;
}


# text : Handles plain text, that are not in tags.
#
#        Removes non-whitelisted characters, and inserts redundant tags
#        between each character.
sub text ($$) {
    my ($self, $text) = @_;

    if (   (@TAG_STACK == 0)
        || ($PERMITTED_TAGS {lc $TAG_STACK [-1]})) {
        printCensored ($text);
    }
}


# start : Handles opening tags.
#
#         Ignores non-whitelisted tags.
sub start ($$) {
    my ($self, $tag, $attr, $attrSeq, $origtext) = @_;

    my $trueTag = $tag;
    $tag =~ s/\/[\s]*$//; # Self-closing tags

    push @TAG_STACK, $tag if $tag eq $trueTag;

    if (defined $PERMITTED_TAGS {lc $tag}) {
        foreach my $currentAttrSeq (@$attrSeq) {
            my $currentAttr = $$attr {$currentAttrSeq};

            if (   (defined $INLINED_ATTRIBUTES {$tag})
                && ($INLINED_ATTRIBUTES {$tag} eq $currentAttrSeq)) {
                printUncensored " <font color='blue'><b>";
                $self->text ($currentAttr);
                printUncensored "</b></font> ";
            }
        }

        if (defined $BLOCKQUOTE_TAGS {$tag}) {
            printUncensored "<" . $BLOCKQUOTE_TAGS {$tag};
        } else {
            printUncensored "<$tag";
        }

        foreach my $currentAttrSeq (@$attrSeq) {
            my $currentAttr = $$attr {$currentAttrSeq};

            my $tagRegex = $PERMITTED_ATTRIBUTES {$currentAttrSeq};
            if ((defined ($tagRegex)) && ($currentAttr =~ /^$tagRegex$/i)) {
                printUncensored " $currentAttrSeq='$currentAttr'";
#            } else {
#                print STDERR "    $tag $currentAttrSeq = $currentAttr\n";
            }
        }

        printUncensored ">";
    } else {
        $REJECTED_TAGS {$tag} ++;
    }   
}


# end : Handles closing tags.
#
#       Ignores non-whitelisted tags.
sub end ($$) {
    my ($self, $tag, $origText) = @_;

    if (defined $PERMITTED_TAGS {lc $tag}) {
        if (defined $BLOCKQUOTE_TAGS {$tag}) {
            printUncensored "</" . $BLOCKQUOTE_TAGS {$tag} . ">";
        } else {
            printUncensored "</$tag>";
        }
    }
        
    pop @TAG_STACK if $TAG_STACK [-1] eq $tag;
}


# stripComments : Saves a copy of standard input to the file, without
#                 the comments.
sub stripComments ($) {
    my ($outFH) = @_;

    my @lines = ();
    my $commentOn = $FALSE;

    while (my $classicLine = <STDIN>) {
        my $line = $classicLine;
        my $commentRemainder = $FALSE;

        while ($line =~ /\Q<!--\E/) {
            $commentRemainder = $TRUE;

            if ($line =~ /\Q-->\E/) {
                $line =~ s/\Q<!--\E.*?\Q-->\E//;
                $commentOn = $FALSE;
            } else {
                $line =~ s/\Q<!--\E.*//;
                $commentOn = $TRUE;
            }
        }

        if ($line =~ /\Q-->\E/) {
            $line =~ s/.*\Q-->\E//;
            $commentOn = $FALSE;
        }

        $line =~ s/\Q$NOP\E//g;
        $line =~ s/\Q$NOP_FOR_CODE\E//g;

        print $outFH $line if ($commentRemainder || ! $commentOn);
    }
    
    close $outFH;
}


# printOrdered : Displays the key/value pairs of a hash, with keys in
#                numeric order.
sub printOrdered ($) {
    my ($hashRef) = @_;

    foreach my $key (sort {$$hashRef {$a} <=> $$hashRef {$b}}
                          keys %$hashRef) {
        print STDERR "    $key " . $$hashRef {$key} . "\n";
    }
}


foreach my $tag (@PERMITTED_TAGS) {
    $PERMITTED_TAGS {$tag} = {} unless defined $PERMITTED_TAGS {$tag};
}
foreach my $tag (@ENTITIES) {
    $tag = '&' . $tag . ';';
    $ENTITIES {join $NOP_FOR_CODE, (split //, $tag)} = $tag;
    $ENTITIES {join $NOP,          (split //, $tag)} = $tag;
}


File::Temp->safe_level (File::Temp::HIGH);
my ($tempFH, $tempFileName) = File::Temp->tempfile ("tempXXXX", UNLINK => 1);

stripComments ($tempFH);

my $p = new InanimateMail;
$p->utf8_mode ($TRUE);
$p->parse_file ($tempFileName);

# print STDERR "Rejected tags:\n";
# printOrdered (\%REJECTED_TAGS);
# print STDERR "\n";

# print STDERR "Rejected chars:\n";
# printOrdered (\%REJECTED_CHARS);
# print STDERR "\n";

